# -*- coding: utf-8 -*-
"""
Created on Fri Sep 20 14:39:19 2019

@author: icruicks
"""

import networkx as nx, numpy as np, pandas as pd
from scipy.sparse import csr_matrix
from sklearn.cluster import SpectralClustering
from sknetwork.clustering import Louvain, modularity 
from scipy.spatial.distance import jaccard
from sklearn.preprocessing import normalize
from sklearn.base import ClusterMixin


class MCLA(ClusterMixin):
    """MCLA+ is cluster ensembling algorithm that works on the meta-cluster graph
    
    MCLA : turn the objects-by-clusterings into a meta-cluster graph of cluster-to-cluster
    by shared objects. Algorithm includes several modifications for different weights
    for the meta-cluster graph as well as the ability to threshold
    
    Parameters
    ----------
    base_clusters : list of array_like or pandas dataframe
        List of cluster labels where each set of labels must be in the same order
        and have shape (n_objects,). Note: standard output from sci-kit learn 
        clustering algorithms will return this format. If pandas dataframe, 
        should be shape (n_objects, n_clusters)
    metaclustering_alg : string
        string to denote how to cluster the object-to-object cluster association
        graph. Options are: 'louvain','spectral','METIS'. Default is 'louvain'
    n_clusters : int
        Number of clusters to find. Used only for spectral and METIS. Default
        is None
    threshold : float
        The threshold for locally pruning the lowest ratio of entries
        in the cluster association matrix. Default is 0
    cluster_comparison : boolean
        How to weight the edges of the meta-cluster graph by their shared objects.
        Default is 'rsc'
    final_clustering_alg : string
        string for how to decide to assign object labels by the meta-clusters. 
        Options are: 'graph_louvain' and 'argmax'. Default is 'argmax'
    
    Returns
    -------
    z : ndarray
        cluster labels of shape (n,).
    """
    
    def __init__(self, metaclustering_alg = 'louvain', n_clusters = None, threshold=0,
                 cluster_comparison = 'rsc', final_clustering_alg='argmax'):
        
        self.metaclustering_alg = metaclustering_alg
        self.n_clusters = n_clusters
        self.tau = threshold
        self.cluster_comparison = cluster_comparison
        self.final_clustering_alg = final_clustering_alg
        
        
    def fit_predict(self, base_clusters):
        
        #create the binary association matrices from the list of clustering
        #labels and calso track where the different clusterings are
        ba_matrices =[]
        idxs = []
        idxs_iter = 0
        
        if isinstance(base_clusters, pd.DataFrame):
            base_clusters = [base_clusters.loc[:,i].to_numpy(dtype=int) for i in base_clusters.columns]
        else:
            base_clusters = [clustering.astype(np.int64) for clustering in base_clusters]
        
        for base_cluster in base_clusters:
            ba_matrix = np.zeros((base_cluster.size, base_cluster.max()+1))
            ba_matrix[np.arange(base_cluster.size), base_cluster] = 1
            ba_matrices.append(ba_matrix)
            idxs.append([idxs_iter, idxs_iter+ba_matrix.shape[1]])
            idxs_iter += ba_matrix.shape[1]
            
        ba_matrix = np.concatenate(ba_matrices, axis=1)
        
        #create the meta-graph of cluster to cluster
        ca_adjacency = np.zeros((ba_matrix.shape[1], ba_matrix.shape[1]))
        weighting = edge_weighting()
        meta_clusterer = meta_alg()
        
        for col_idx in range(ba_matrix.shape[1]):
            for idx in idxs:
                if idx[0] <= col_idx < idx[1]:
                    pass
                else:
                    for alt_idx in range(idx[0], idx[1]):
                        if ca_adjacency[col_idx, alt_idx] <= 0:
                            ca_adjacency[col_idx, alt_idx] = weighting.weight(ba_matrix[:,col_idx], ba_matrix[:,alt_idx], self.cluster_comparison)
                            # ca_adjacency[alt_idx, col_idx] = ca_adjacency[col_idx, alt_idx]
            
        ca_adjacency = self._prune(ca_adjacency)
        self.meta_cluster_adjacency = ca_adjacency.copy()
        meta_clusters = meta_clusterer.cluster(ca_adjacency, self.n_clusters, self.metaclustering_alg)+1
        aug_matrix = np.int64(np.multiply(ba_matrix, meta_clusters.T))
        
        # produce the final clustering of the objects by the different meta-clusters
        #that they fall int0.
        final_clusterer = final_alg()
        subgroup_counts = self._bincount2D_vectorized(aug_matrix)
        normalized_counts = normalize(subgroup_counts, norm='l1', axis=0)
        return final_clusterer.cluster(normalized_counts, self.final_clustering_alg)
    
    def _bincount2D_vectorized(self, a):   
        '''
        Internal helper function to get the counts of the number of clusters that
        each object has that falls into each of the found meta-clusters
        '''
        N = a.max()+1
        a_offs = a + np.arange(a.shape[0])[:,None]*N
        return np.bincount(a_offs.ravel(), minlength=a.shape[0]*N).reshape(-1,N)[:,1:]
        
    def _prune(self, adjacency):
        '''
        Internal function for global-local thresholding
        '''
        thresh = np.apply_along_axis(lambda row: np.quantile(row[np.nonzero(row)], self.tau), 1, adjacency)
        mins = np.apply_along_axis(lambda row: np.min(row[np.nonzero(row)]), 1, adjacency)
        global_thresh = np.quantile(adjacency[np.nonzero(adjacency)], self.tau)
        thresh[mins > global_thresh] = 0
        adjacency[adjacency < thresh] = 0
        # adjacency[adjacency < global_thresh] = 0
        return adjacency



class LWMC(ClusterMixin):
    """Locally-Weighted Neta-Clustering cluster ensembling algorithm 
    that works on the meta-cluster graph
    
    LWMC: turn the objects-by-clusterings into a meta-cluster graph of cluster-to-cluster
    by shared objects. Algorithm adapts traditional MCLA to cluster to cluster 
    weightings by information-theorteic measures
    
    Parameters
    ----------
    base_clusters : list of array_like or pandas dataframe
        List of cluster labels where each set of labels must be in the same order
        and have shape (n_objects,). Note: standard output from sci-kit learn 
        clustering algorithms will return this format. If pandas dataframe, 
        should be shape (n_objects, n_clusters)
    metaclustering_alg : string
        string to denote how to cluster the object-to-object cluster association
        graph. Options are: 'louvain','spectral','METIS'. Default is 'louvain'
    n_clusters : int
        Number of clusters to find. Used only for spectral and METIS. Default
        is None
    theta : float
        Controls the impact of the local weighting. Default is 0.5
    cluster_comparison : boolean
        How to weight the edges of the meta-cluster graph by their shared objects.
        Default is 'rsc'
    final_clustering_alg : string
        string for how to decide to assign object labels by the meta-clusters. 
        Options are: 'graph_louvain' and 'argmax'. Default is 'argmax'
    
    Returns
    -------
    z : ndarray
        cluster labels of shape (n,).
    """
    
    def __init__(self, metaclustering_alg = 'louvain', n_clusters = None, 
                 cluster_comparison = 'jaccard', final_clustering_alg='argmax',
                 theta=0.5):
        
        self.metaclustering_alg = metaclustering_alg
        self.n_clusters = n_clusters
        self.cluster_comparison = cluster_comparison
        self.final_clustering_alg = final_clustering_alg
        self.theta = theta
        
        
    def fit_predict(self, base_clusters):
        ba_matrices =[]
        idxs = []
        idxs_iter = 0
        
        if isinstance(base_clusters, pd.DataFrame):
            base_clusters = [base_clusters.loc[:,i].to_numpy(dtype=int) for i in base_clusters.columns]
        else:
            base_clusters = [clustering.astype(np.int64) for clustering in base_clusters]
        
        for base_cluster in base_clusters:
            ba_matrix = np.zeros((base_cluster.size, base_cluster.max()+1))
            ba_matrix[np.arange(base_cluster.size), base_cluster] = 1
            ba_matrices.append(ba_matrix)
            idxs.append([idxs_iter, idxs_iter+ba_matrix.shape[1]])
            idxs_iter += ba_matrix.shape[1]
            
        ba_matrix = np.concatenate(ba_matrices, axis=1)
        
        ca_adjacency = np.zeros((ba_matrix.shape[1], ba_matrix.shape[1]))
        H_matrix = np.zeros((ba_matrix.shape[1], len(ba_matrices)))
        weighting = edge_weighting()
        meta_clusterer = meta_alg()
        final_clusterer = final_alg()        
        
        for col_idx in range(ba_matrix.shape[1]):
            for mode in range(len(idxs)):
                idx = idxs[mode]
                if idx[0] <= col_idx < idx[1]:
                    pass
                else:
                    H_m = []
                    for alt_idx in range(idx[0], idx[1]):
                        C_i = ba_matrix[:,col_idx]
                        C_j = ba_matrix[:,alt_idx]
                        ca_adjacency[col_idx, alt_idx] = weighting.weight(C_i, C_j, self.cluster_comparison)
                        ca_adjacency[alt_idx, col_idx] = ca_adjacency[col_idx, alt_idx]
                        P_ij = np.dot(C_i, C_j)/ np.count_nonzero(C_i)
                        if P_ij > 0:
                            H_m.append(P_ij * np.log2(P_ij))
                        else:
                            H_m.append(0)
                    H_matrix[col_idx, mode] = -1*np.sum(H_m)
        
        ECI = np.exp(-1* np.sum(H_matrix, axis=1)/(self.theta*len(ba_matrices)))
        
        self.meta_cluster_graph = nx.from_numpy_array(ca_adjacency)
        meta_clusters = meta_clusterer.cluster(ca_adjacency, self.n_clusters, self.metaclustering_alg)+1
        aug_matrix = np.int64(np.multiply(ba_matrix, meta_clusters.T))
        weighted_subgroup_counts = self._get_weighted_subgroup_counts(aug_matrix, ECI)
        return final_clusterer.cluster(weighted_subgroup_counts, self.final_clustering_alg)
    
    def _get_weighted_subgroup_counts(self, a, cluster_weights):    
        N = a.max()+1
        weighted_counts = np.zeros((a.shape[0], N-1))
        for label in range(1,N):
            label_counts = (a==label) * cluster_weights
            normalization = np.count_nonzero(np.sum(a == label, axis=0) >0)
            weighted_counts[:,label-1] = np.sum(label_counts, axis=1)/normalization
        return weighted_counts

                
class edge_weighting:
    '''
    Factory class for getting the cluster-to-cluster edge weights, based on
    shared objects
    '''
    def weight(self, col_1, col_2, format):
        weighting = self._get_weighting(format)
        return weighting(col_1, col_2)
    
    def _get_weighting(self, format):
        if format == 'jaccard':
            return self._jc
        elif format == 'hpi':
            return self._hpi
        elif format == 'rsc':
            return self._rsc
        elif format == 'sorensen':
            return self._sorensen
        elif format == 'salton':
            return self._salton
        elif format == 'lhn':
            return self._lhn
        elif format == 'mutual_information':
            return self._mi
        elif format == 'intersection_probability':
            return self._intersection_prob
        else:
            raise ValueError(format)
        
    def _jc(self, col_1, col_2):
        return 1- jaccard(col_1, col_2)
    
    def _hpi(self, col_1, col_2):
        return np.dot(col_1, col_2)/ np.minimum(np.sum(col_1), np.sum(col_2))
    
    def _rsc(self, col_1, col_2):
        n= col_1.shape[0]
        size_1 = np.sum(col_1)
        size_2 = np.sum(col_2)
        numerator = np.dot(col_1, col_2) - (size_1*size_2)/n
        denominator = np.sqrt(size_1*size_2*(1-size_1/n)*(1-size_2/n))
        return np.maximum(numerator/denominator, 0)
    
    def _sorensen(self, col_1, col_2):
        return (2*np.dot(col_1, col_2))/(np.sum(col_1)+np.sum(col_2))
    
    def _salton(self, col_1, col_2):
        return np.dot(col_1, col_2)/np.sqrt(np.sum(col_1)*np.sum(col_2))
    
    def _lhn(self, col_1, col_2):
        return np.dot(col_1, col_2)/(np.sum(col_1)*np.sum(col_2))
    
    def _mi(self, col_1, col_2):
        n= col_1.shape[0]
        p_ij = np.dot(col_1, col_2)/n
        p_i = np.sum(col_1)/n
        p_j = np.sum(col_2)/n
        if p_ij >0:
            return p_ij*np.log(p_ij/(p_i * p_j))
        else:
            return 0
        
    def _intersection_prob(self, col_1, col_2):
        return np.dot(col_1, col_2)/np.count_nonzero(col_1)
    
 
class meta_alg:
    '''
    Factory class for clustering the meta-graphs
    '''
    def cluster(self, adj, n_clusters, format):
        clustering = self._get_clustering(format)
        return clustering(adj, n_clusters)
    
    def _get_clustering(self, format):
        if format == 'louvain':
            return self._louvain
        elif format == 'spectral':
            return self._spectral
        elif format == 'METIS':
            return self._metis
        else:
            raise ValueError(format)
            
    def _louvain(self, ca_adjacency, n_clusters):
        cluster_iterations = [Louvain(random_state=int(rand_state), shuffle_nodes=True) for rand_state in np.random.randint(1,high=100000,size=10)]
        parts = [clstr.fit(ca_adjacency).labels_ for clstr in cluster_iterations]
        modularities = [modularity(ca_adjacency, part) for part in parts]
        return parts[np.argmax(modularities)]
    
    def _spectral(self, ca_adjacency, n_clusters):
        clstr = SpectralClustering(n_clusters = n_clusters, affinity='precomputed', n_jobs=-1)
        return clstr.fit_predict(ca_adjacency)+1
    
    def _metis(self, ca_adjacency, n_clusters):
        try:
            import nxmetis
        except:
            print("METIS is not able to import.")
        binary_adjacency = ca_adjacency.copy()
        binary_adjacency[binary_adjacency>0]=1
        ca_graph = nx.from_numpy_array(binary_adjacency)
        _, temp_clusters = nxmetis.partition(ca_graph, n_clusters)
        meta_clusters = np.zeros(ca_adjacency.shape[0])
        for i in range(len(temp_clusters)):
            for j in range(len(temp_clusters[i])):
                meta_clusters[temp_clusters[i][j]] = i
        return meta_clusters+1
        
        
class final_alg:
    '''
    Factory class for doing the final clustering of the objects by meta-clusters
    '''
    def cluster(self, counts, format):
        clustering = self._get_clustering(format)
        return clustering(counts)
    
    def _get_clustering(self, format):
        if format == 'argmax':
            return self._argmax
        elif format == 'graph_louvain':
            return self._graph_louvain
        else:
            raise ValueError(format)
            
    def _argmax(self, counts):
        return np.argmax(counts, axis=1)
    
    def _graph_louvain(self, counts):
        final_graph = csr_matrix(np.matmul(counts, counts.T))
        cluster_iterations = [Louvain(random_state=int(rand_state), shuffle_nodes=True) for rand_state in np.random.randint(1,high=100000,size=10)]
        parts = [clstr.fit(final_graph).labels_ for clstr in cluster_iterations]
        modularities = [modularity(final_graph, part) for part in parts]
        return parts[np.argmax(modularities)]
        
        
        
        
        
        
        
        
        